import { EngineComponent } from '../types/engine-component';
import { assetManager } from '../assets/manager';
import { World } from '../gameobjects/world.object';
import { instancableGameObjects } from '../gameobjects';
import { environmentDefaults } from '../defaults/environment';
/**
 * Game level management component. This component provides the World.
 *
 * EnvironmentComponent must be initialized before this component.
 * @listens change Applies changes to objects
 * @listens remove Removes objects from scene
 * @listens reparent Reparents object
 */
export class LevelComponent extends EngineComponent {
    constructor(renderer, events) {
        super(renderer, events);
        this.renderer = renderer;
        this.events = events;
        this.name = LevelComponent.name;
        this.world = new World();
    }
    initialize() {
        this.renderer.scene.add(this.world);
        this.environment = this.renderer.scene.getObjectByName('Environment');
        this.cleanUpEvents = this.bindEvents();
    }
    update(delta) { }
    dispose() {
        this.cleanUpEvents?.call(this);
    }
    getSceneTree() {
        return {
            world: this.world,
            environment: this.environment,
        };
    }
    createObject(object, setParent, skipJoin = false) {
        const parent = setParent || this.world;
        const ObjectType = instancableGameObjects[object];
        if (!ObjectType)
            return;
        const newObject = new ObjectType();
        parent.add(newObject);
        if (!skipJoin)
            this.events.emit('sceneJoin', newObject);
        return newObject;
    }
    deserializeObject(root, setParent) {
        const parent = setParent || this.world;
        if (root.objectType === 'World') {
            root.children.forEach((entry) => this.recursiveCreate(entry, parent));
            return parent;
        }
        return this.recursiveCreate(root, parent);
    }
    serializeLevelSave(name) {
        const world = this.world.serialize();
        const environment = this.environment.serialize();
        const assets = assetManager.serialize();
        return {
            name,
            world,
            environment,
            assets,
        };
    }
    async deserializeLevelSave(save) {
        // Reset the world
        this.events.emit('reset');
        // Load all assets
        await assetManager.loadAll(save.assets);
        // Load environment
        this.environment.deserialize(save.environment);
        this.events.emit('setEnvironment', this.environment);
        // Load world
        this.deserializeObject(save.world);
        this.world.deserialize(save.world);
        this.events.emit('loadComplete');
    }
    recursiveCreate(entry, setParent) {
        const parent = setParent || this.world;
        const newObject = this.createObject(entry.objectType, parent, true);
        newObject?.deserialize(entry);
        entry.children.forEach((child) => this.recursiveCreate(child, newObject));
        return newObject;
    }
    bindEvents() {
        const changeEvent = (event) => {
            if (event.applied || !event.object)
                return;
            const prop = event.object[event.property];
            if (prop?.copy)
                prop.copy(event.value);
            else
                event.object[event.property] = event.value;
        };
        const removeEvent = (event) => {
            if (event.applied || !event.object)
                return;
            if (event.object.virtual)
                return;
            if (Array.isArray(event.object)) {
                event.object.forEach((object) => object.removeFromParent());
                return;
            }
            event.object.removeFromParent();
            this.events.emit('sceneLeave', event.object);
        };
        const reparentEvent = (event) => {
            if (event.applied || !event.object || !event.parent)
                return;
            if (Array.isArray(event.object)) {
                event.object.forEach((object) => {
                    event.parent.attach(object);
                    this.events.emit('change', {
                        object: object,
                        property: 'parent',
                        value: event.parent,
                        applied: true,
                    });
                });
                return;
            }
            event.parent.attach(event.object);
            this.events.emit('change', {
                object: event.object,
                property: 'parent',
                value: event.parent,
                applied: true,
            });
        };
        const instanceEvent = (event) => {
            const parent = typeof event.parent === 'string'
                ? this.world.getObjectByProperty('uuid', event.parent)
                : event.parent;
            const object = this.createObject(event.type, parent, !!event.data);
            if (event.data && object) {
                object.deserialize(event.data);
                this.events.emit('sceneJoin', object);
            }
        };
        const resetEvent = () => {
            this.world.clear();
            assetManager.freeAll();
            this.events.emit('setEnvironment', environmentDefaults);
            Object.assign(this.environment, environmentDefaults);
        };
        const queueFreeEvent = (obj) => {
            const object = typeof obj === 'string'
                ? this.world.getObjectByProperty('uuid', obj)
                : obj;
            if (!object)
                return;
            this.events.emit('remove', { object });
        };
        this.events.addListener('change', changeEvent);
        this.events.addListener('remove', removeEvent);
        this.events.addListener('reparent', reparentEvent);
        this.events.addListener('instance', instanceEvent);
        this.events.addListener('reset', resetEvent);
        this.events.addListener('queueFree', queueFreeEvent);
        return () => {
            this.events.removeEventListener('change', changeEvent);
            this.events.removeEventListener('remove', removeEvent);
            this.events.removeEventListener('reparent', reparentEvent);
            this.events.removeEventListener('instance', instanceEvent);
            this.events.removeEventListener('reset', resetEvent);
            this.events.removeEventListener('queueFree', queueFreeEvent);
        };
    }
}
