import { Raycaster, Vector2 } from 'three';
import { EngineComponent } from '../types/engine-component';
/**
 * Manage mouse and object picking from screen.
 *
 * ViewportComponent and LevelComponent must be initialized before this component.
 */
export class MouseComponent extends EngineComponent {
    constructor(renderer, events) {
        super(renderer, events);
        this.renderer = renderer;
        this.events = events;
        this.name = MouseComponent.name;
        this.mouseButtons = [false, false, false];
        this.mouseButtonsLast = [false, false, false];
        this.mousePosition = new Vector2(0, 0);
        this.mousepositionNDC = new Vector2(0, 0);
        this.mousePositionLast = new Vector2(0, 0);
        this.mousepositionNDCLast = new Vector2(0, 0);
        this.canvasOffset = new Vector2(0, 0);
        this.ray = new Raycaster();
    }
    get canvas() {
        return this.renderer.renderer.domElement;
    }
    initialize() {
        this.world = this.renderer.scene.getObjectByName('World');
        this.cleanUpEvents = this.bindEvents();
    }
    update(delta) {
        this.ray.setFromCamera(this.mousepositionNDC, this.renderer.camera);
        this.mouseButtonsLast = [...this.mouseButtons];
    }
    dispose() {
        this.cleanUpEvents?.call(this);
    }
    bindEvents() {
        const mouseDown = (ev) => {
            const [object, ...rest] = this.ray.intersectObjects(this.world.children, true);
            this.mouseButtons[ev.button] = true;
            this.events.emit('mouseDown', {
                position: this.mousePosition,
                positionNDC: this.mousepositionNDC,
                button: ev.button,
                target: object,
                shift: ev.shiftKey,
                control: ev.ctrlKey,
                alt: ev.altKey,
                raw: ev,
                otherTargets: rest,
            });
        };
        const mouseUp = (ev) => {
            const [object, ...rest] = this.ray.intersectObjects(this.world.children, true);
            this.mouseButtons[ev.button] = false;
            this.events.emit('mouseUp', {
                position: this.mousePosition,
                positionNDC: this.mousepositionNDC,
                button: ev.button,
                target: object,
                shift: ev.shiftKey,
                control: ev.ctrlKey,
                alt: ev.altKey,
                raw: ev,
                otherTargets: rest,
            });
        };
        const mouseMove = (ev) => {
            this.mousePositionLast = this.mousePosition.clone();
            this.mousepositionNDCLast = this.mousepositionNDC.clone();
            this.mousePosition.set(ev.clientX, ev.clientY).sub(this.canvasOffset);
            this.mousepositionNDC.set((this.mousePosition.x / this.renderer.resolution.x) * 2 - 1, -(this.mousePosition.y / this.renderer.resolution.y) * 2 + 1);
            this.events.emit('mouseMove', {
                position: this.mousePosition,
                positionNDC: this.mousepositionNDC,
                offset: this.mousePositionLast.clone().sub(this.mousePosition),
                offsetGL: this.mousepositionNDCLast.clone().sub(this.mousepositionNDC),
                raw: ev,
            });
        };
        const contextMenu = (ev) => ev.preventDefault();
        const onResize = () => {
            const calculate = this.canvas.getBoundingClientRect();
            this.canvasOffset.set(calculate.left, calculate.top);
        };
        this.canvas.addEventListener('mousedown', mouseDown);
        this.canvas.addEventListener('mousemove', mouseMove);
        this.canvas.addEventListener('mouseup', mouseUp);
        this.canvas.addEventListener('contextmenu', contextMenu);
        this.events.addListener('resize', onResize);
        onResize();
        return () => {
            this.canvas.removeEventListener('mousedown', mouseDown);
            this.canvas.removeEventListener('mousemove', mouseMove);
            this.canvas.removeEventListener('mouseup', mouseUp);
            this.canvas.removeEventListener('contextmenu', contextMenu);
            this.events.removeEventListener('resize', onResize);
        };
    }
}
