// Contains highly modified code from
// https://github.com/mrdoob/three.js/blob/master/examples/jsm/controls/FirstPersonControls.js
// https://github.com/mrdoob/three.js/blob/master/examples/jsm/controls/OrbitControls.js
// MIT Copyright © 2010-2023 three.js authors
import { Euler, EventDispatcher, Vector3 } from 'three';
const _changeEvent = { type: 'change' };
const _PI_2 = Math.PI / 2;
/**
 * Freefly camera controls.
 * WASD + QE to move around.
 * Holding Right Mouse button rotates the view.
 * Holding Middle Mouse button pans the view.
 * Shift slows down movement.
 */
class CameraControls extends EventDispatcher {
    constructor(camera, domElement) {
        super();
        this.camera = camera;
        this.domElement = domElement;
        this.cameraMoving = false;
        this.cameraPanning = false;
        // Set to constrain the pitch of the camera
        // Range is 0 to Math.PI radians
        this.minPolarAngle = 0; // radians
        this.maxPolarAngle = Math.PI; // radians
        this.pointerSpeed = 1.5;
        this.panSpeed = 1;
        this.movementSpeed = 16;
        this.shiftMultiplier = 0.45;
        this.zoomScale = 100;
        this.screenSpacePanning = true;
        this.rotation = new Euler(0, 0, 0, 'YXZ');
        this.look = new Vector3();
        this.boundOnMouseMove = this.onMouseMove.bind(this);
        this.boundOnMouseDown = this.onMouseDown.bind(this);
        this.boundOnMouseUp = this.onMouseUp.bind(this);
        this.boundOnKeyDown = this.onKeyDown.bind(this);
        this.boundOnKeyUp = this.onKeyUp.bind(this);
        this.boundOnScroll = this.onScroll.bind(this);
        this.movement = {
            forward: 0,
            backward: 0,
            left: 0,
            right: 0,
            up: 0,
            down: 0,
            multiplier: 1,
        };
        this.connect();
    }
    connect() {
        this.domElement.addEventListener('mousemove', this.boundOnMouseMove);
        this.domElement.addEventListener('mousedown', this.boundOnMouseDown);
        this.domElement.addEventListener('mouseup', this.boundOnMouseUp);
        this.domElement.addEventListener('wheel', this.boundOnScroll, {
            passive: true,
        });
        window.addEventListener('keydown', this.boundOnKeyDown);
        window.addEventListener('keyup', this.boundOnKeyUp);
    }
    disconnect() {
        this.domElement.removeEventListener('mousemove', this.boundOnMouseMove);
        this.domElement.removeEventListener('mousedown', this.boundOnMouseDown);
        this.domElement.removeEventListener('mouseup', this.boundOnMouseUp);
        this.domElement.removeEventListener('wheel', this.boundOnScroll);
        window.removeEventListener('keydown', this.boundOnKeyDown);
        window.removeEventListener('keyup', this.boundOnKeyUp);
    }
    dispose() {
        this.disconnect();
    }
    update(dt) {
        if (this.movement.forward !== 0) {
            this.moveForward(this.movement.forward * dt * this.movement.multiplier);
        }
        if (this.movement.backward !== 0) {
            this.moveForward(-this.movement.backward * dt * this.movement.multiplier);
        }
        if (this.movement.right !== 0) {
            this.moveRight(this.movement.right * dt * this.movement.multiplier);
        }
        if (this.movement.left !== 0) {
            this.moveRight(-this.movement.left * dt * this.movement.multiplier);
        }
        if (this.movement.up !== 0) {
            this.moveUp(this.movement.up * dt * this.movement.multiplier);
        }
        if (this.movement.down !== 0) {
            this.moveUp(-this.movement.down * dt * this.movement.multiplier);
        }
    }
    getDirection(v) {
        return v.set(0, 0, -1).applyQuaternion(this.camera.quaternion);
    }
    moveForward(distance) {
        const camera = this.camera;
        this.rotation.setFromQuaternion(camera.quaternion);
        this.look.setFromMatrixColumn(camera.matrix, 0);
        this.look.crossVectors(camera.up.clone().applyEuler(this.rotation), this.look);
        camera.position.addScaledVector(this.look, distance);
    }
    moveUp(distance) {
        const camera = this.camera;
        this.rotation.setFromQuaternion(camera.quaternion);
        this.look.copy(camera.up.clone().applyEuler(this.rotation));
        camera.position.addScaledVector(this.look, distance);
    }
    moveRight(distance) {
        const camera = this.camera;
        this.look.setFromMatrixColumn(camera.matrix, 0);
        camera.position.addScaledVector(this.look, distance);
    }
    onMouseDown(event) {
        if (event.button === 1)
            this.cameraPanning = true;
        if (event.button === 2)
            this.cameraMoving = true;
    }
    onMouseUp(event) {
        if (event.button === 1)
            this.cameraPanning = false;
        if (event.button === 2)
            this.cameraMoving = false;
    }
    onScroll(event) {
        if (event.deltaY < 0) {
            this.moveForward(this.movementSpeed * 0.002 * this.zoomScale);
        }
        else if (event.deltaY > 0) {
            this.moveForward(-this.movementSpeed * 0.002 * this.zoomScale);
        }
    }
    rotate(event) {
        const movementX = event.movementX || 0;
        const movementY = event.movementY || 0;
        const camera = this.camera;
        this.rotation.setFromQuaternion(camera.quaternion);
        this.rotation.y -= movementX * 0.002 * this.pointerSpeed;
        this.rotation.x -= movementY * 0.002 * this.pointerSpeed;
        this.rotation.x = Math.max(_PI_2 - this.maxPolarAngle, Math.min(_PI_2 - this.minPolarAngle, this.rotation.x));
        camera.quaternion.setFromEuler(this.rotation);
    }
    pan(event) {
        const movementX = event.movementX || 0;
        const movementY = event.movementY || 0;
        const camera = this.camera;
        this.look.setFromMatrixColumn(camera.matrix, 0);
        camera.position.addScaledVector(this.look, -movementX * 0.02 * this.panSpeed);
        if (this.screenSpacePanning) {
            this.look.setFromMatrixColumn(camera.matrix, 1);
        }
        else {
            this.look.setFromMatrixColumn(camera.matrix, 0);
            this.look.crossVectors(camera.up, this.look);
        }
        camera.position.addScaledVector(this.look, movementY * 0.02 * this.panSpeed);
    }
    onMouseMove(event) {
        if (this.cameraMoving === false && this.cameraPanning === false)
            return;
        this.cameraMoving && this.rotate(event);
        this.cameraPanning && this.pan(event);
        this.dispatchEvent(_changeEvent);
    }
    onKeyDown(event) {
        if (!event.target || event.target.tagName !== 'BODY')
            return;
        if (event.altKey || event.ctrlKey)
            return;
        switch (event.code) {
            case 'ShiftLeft':
                this.movement.multiplier = this.shiftMultiplier;
                break;
            case 'KeyE':
                this.movement.up = this.movementSpeed;
                break;
            case 'KeyQ':
                this.movement.down = this.movementSpeed;
                break;
            case 'KeyW':
                this.movement.forward = this.movementSpeed;
                break;
            case 'KeyS':
                this.movement.backward = this.movementSpeed;
                break;
            case 'KeyA':
                this.movement.left = this.movementSpeed;
                break;
            case 'KeyD':
                this.movement.right = this.movementSpeed;
                break;
        }
    }
    onKeyUp(event) {
        if (!event.target || event.target.tagName !== 'BODY')
            return;
        if (event.altKey || event.ctrlKey)
            return;
        switch (event.code) {
            case 'ShiftLeft':
                this.movement.multiplier = 1;
                break;
            case 'KeyE':
                this.movement.up = 0;
                break;
            case 'KeyQ':
                this.movement.down = 0;
                break;
            case 'KeyW':
                this.movement.forward = 0;
                break;
            case 'KeyS':
                this.movement.backward = 0;
                break;
            case 'KeyA':
                this.movement.left = 0;
                break;
            case 'KeyD':
                this.movement.right = 0;
                break;
        }
    }
}
export { CameraControls };
