import { Packet, PacketType } from '.';
export class GameSocket {
    constructor(events) {
        this.events = events;
    }
    track(world) {
        this.world = world;
    }
    connect(host, id, token) {
        this.host = host;
        this.playerId = id;
        this.token = token;
        this.ws = new WebSocket(host);
        this.ws.binaryType = 'arraybuffer';
        this.bindSocket(this.ws);
        console.log('connected to', this.host);
        this.cleanUpEvents = this.bindEvents();
    }
    disconnect() {
        this.ws?.close();
        this.events.emit('serverDisconnect');
        console.log('disconnected from', this.host);
        this.cleanUpEvents?.();
    }
    dispose() {
        this.disconnect();
    }
    onOpen() {
        const packet = new Packet(PacketType.AUTH)
            .write(this.playerId, String)
            .write(this.token, String)
            .write(0, 'uint8')
            .toBuffer();
        this.ws?.send(packet);
    }
    handlePacket(incoming) {
        switch (incoming.packet) {
            case PacketType.ERROR: {
                const error = incoming.read(String);
                alert(error);
                break;
            }
            case PacketType.STREAM_START:
                console.log('starting world loading');
                break;
            case PacketType.STREAM_OBJECT: {
                const objectUUID = incoming.read(String);
                const parentUUID = incoming.read(String);
                const objectType = incoming.read(String);
                const objectData = JSON.parse(incoming.read(String));
                this.events.emit('instance', {
                    type: objectType,
                    parent: parentUUID,
                    data: objectData,
                });
                break;
            }
            case PacketType.STREAM_DESTROY: {
                const objectUUID = incoming.read(String);
                this.events.emit('queueFree', objectUUID);
                break;
            }
            case PacketType.STREAM_TRANSFORM: {
                const objectUUID = incoming.read(String);
                const objectPos = incoming.read('vec3');
                const objectQuat = incoming.read('quat');
                const objectLinvel = incoming.read('vec3');
                const objectAngvel = incoming.read('vec3');
                this.events.emit('serverTransform', {
                    object: objectUUID,
                    position: objectPos,
                    quaternion: objectQuat,
                    velocity: objectLinvel,
                    angularVelocity: objectAngvel,
                });
                break;
            }
            case PacketType.STREAM_FINISH: {
                this.events.emit('loadComplete');
                break;
            }
            case PacketType.PLAYER_LIST: {
                const playerCount = incoming.read('uint32');
                const players = Array.from({ length: playerCount }, () => null).map(() => incoming.read(String)?.split(':'));
                console.log('player list', players);
                break;
            }
            case PacketType.PLAYER_JOIN: {
                const playerId = incoming.read(String);
                const playerName = incoming.read(String);
                console.log('player joined', playerId, playerName);
                break;
            }
            case PacketType.PLAYER_QUIT: {
                const playerId = incoming.read(String);
                const playerName = incoming.read(String);
                console.log('player quit', playerId, playerName);
                break;
            }
            case PacketType.PLAYER_CHARACTER: {
                const playerId = incoming.read(String);
                const playerName = incoming.read(String);
                const position = incoming.read('vec3');
                const rotation = incoming.read('quat');
                const objectUUID = incoming.read(String);
                this.events.emit('spawnCharacter', {
                    playerId,
                    playerName,
                    position,
                    rotation,
                    objectUUID,
                });
                break;
            }
            case PacketType.PLAYER_MOVEMENT: {
                const playerUUID = incoming.read(String);
                const playerVelocity = incoming.read('vec3');
                const playerLookAt = incoming.read('vec3');
                const playerJump = incoming.read('bool');
                this.events.emit('moveCharacter', {
                    playerId: playerUUID,
                    velocity: playerVelocity,
                    lookAt: playerLookAt,
                    jump: playerJump,
                });
                break;
            }
        }
    }
    bindSocket(ws) {
        ws.addEventListener('message', (event) => {
            const packet = Packet.from(Buffer.from(event.data));
            this.handlePacket(packet);
        });
        ws.addEventListener('close', (event) => {
            this.disconnect();
        });
        ws.addEventListener('error', (event) => { });
        ws.addEventListener('open', (event) => {
            this.onOpen();
        });
    }
    bindEvents() {
        const sendPlayerEvent = (event) => {
            if (this.ws?.readyState !== WebSocket.OPEN)
                return;
            this.ws.send(new Packet(PacketType.PLAYER_MOVEMENT)
                .write(event.playerId, String)
                .write(event.velocity, 'vec3')
                .write(event.lookAt, 'vec3')
                .write(event.jump, 'bool')
                .toBuffer());
        };
        const onBeforeUnload = () => {
            this.disconnect();
        };
        this.events.on('sendPlayer', sendPlayerEvent);
        window.addEventListener('beforeunload', onBeforeUnload);
        return () => {
            this.events.removeEventListener('sendPlayer', sendPlayerEvent);
            window.removeEventListener('beforeunload', onBeforeUnload);
        };
    }
}
