import { CubeTexture, CubeTextureLoader, Texture, TextureLoader } from 'three';
import { Asset, AssetType, AssetsEvents } from '../types/asset';
import { EventEmitter } from '../utils/events';
import { GLTF, GLTFLoader } from 'three/addons/loaders/GLTFLoader.js';

export class AssetManagerFactory extends EventEmitter<AssetsEvents> {
  public assets: Asset[] = [];
  public texureLoader = new TextureLoader();
  public cubeTextureLoader = new CubeTextureLoader();
  public gltfLoader = new GLTFLoader();

  constructor() {
    super();
    this.on('load', (input) =>
      Array.isArray(input) ? this.loadAll(input) : this.load(input)
    );
  }

  /**
   * Get an asset by path, this is an unique identifier
   * @param path Asset path
   */
  getAssetByPath(path: string) {
    return this.assets.find((entry) => entry.path === path);
  }

  /**
   * Create a new texture asset
   * @param data Data URI
   * @param name Asset name
   * @param type Texture type
   */
  async createAsset(data: string, name: string, type: AssetType = 'Texture') {
    const asset: Asset = {
      name,
      type,
      data,
    };

    await this.load(asset);

    asset.path = `fblxassetid:${(asset.texture || asset.object)!.uuid}`;
    this.assets.push(asset);

    return asset;
  }

  /**
   * Load asset into a Texture or Model
   * @param asset Remote or local asset data
   */
  async load(asset: Asset) {
    this.emit('loadStart', asset);

    if (asset.type === 'Mesh') {
      const data = await this.loadMeshData(
        asset.remote ? asset.path : asset.data,
        asset.name
      ).catch((error) => {
        this.emit('loadError', error);
        throw error;
      });

      asset.animations = data.animations;
      asset.object = data.scene;

      this.emit('loadComplete', asset);

      return asset;
    }

    return (
      asset.type === 'Texture'
        ? this.loadTextureData(
            asset.remote ? asset.path : asset.data,
            asset.name
          )
        : this.loadCubeTexture(
            asset.remote ? asset.path : asset.data,
            asset.name
          )
    )
      .then((texture) => {
        asset.texture = texture;
        this.emit('loadComplete', asset);
        return asset;
      })
      .catch((error) => {
        this.emit('loadError', error);
        throw error;
      });
  }

  /**
   * Load assets into Textures
   * @param assets Remote or local asset data
   */
  async loadAll(assets: Asset[]) {
    const loaded = await Promise.allSettled(
      assets.map((item) => this.load(item))
    );

    loaded
      .filter((entry) => entry.status === 'rejected')
      .forEach((error) => console.error('Failed loading asset', error));

    this.assets.push(
      ...loaded
        .filter((entry) => entry.status === 'fulfilled')
        .map((fulfilled) => (fulfilled as PromiseFulfilledResult<Asset>).value)
    );
  }

  /**
   * Serialize for save file without Texture data
   */
  serialize(): Asset[] {
    return this.assets.map((entry) => ({
      name: entry.name,
      path: entry.path,
      data: !entry.remote ? entry.data : undefined,
      type: entry.type,
    }));
  }

  /**
   * Free all assets from memory
   */
  freeAll() {
    this.assets.length = 0;
  }

  /**
   * Load texture
   * @param path Path
   * @param name Texture name
   */
  async loadTextureData(path: string, name?: string) {
    return new Promise<Texture>((resolve, reject) => {
      this.texureLoader.load(
        path,
        (texture) => {
          texture.name = name || 'Texture';
          resolve(texture);
        },
        undefined,
        reject
      );
    });
  }

  /**
   * Load cube texture loader
   * @param path pos-x, neg-x, pos-y, neg-y, pos-z, neg-z
   * @param name Texture name
   */
  async loadCubeTexture(path: string[], name?: string) {
    return new Promise<CubeTexture>((resolve, reject) => {
      this.cubeTextureLoader.load(
        path,
        (texture) => {
          texture.name = name || 'Texture';
          resolve(texture);
        },
        undefined,
        reject
      );
    });
  }

  async loadMeshData(path: string, name?: string) {
    return new Promise<GLTF>((resolve, reject) => {
      this.gltfLoader.load(path, (entry) => resolve(entry), undefined, reject);
    });
  }

  async loadJsonData(path: string) {
    return fetch(path).then((res) => res.json());
  }
}

export const assetManager = new AssetManagerFactory();
