// Contains highly modified code from
// https://github.com/mrdoob/three.js/blob/master/examples/jsm/controls/FirstPersonControls.js
// https://github.com/mrdoob/three.js/blob/master/examples/jsm/controls/OrbitControls.js
// MIT Copyright © 2010-2023 three.js authors

import { Euler, EventDispatcher, Object3D, Vector3 } from 'three';

const _changeEvent = { type: 'change' };

const _PI_2 = Math.PI / 2;

/**
 * Freefly camera controls.
 * WASD + QE to move around.
 * Holding Right Mouse button rotates the view.
 * Holding Middle Mouse button pans the view.
 * Shift slows down movement.
 */
class CameraControls extends EventDispatcher {
  private cameraMoving = false;
  private cameraPanning = false;

  // Set to constrain the pitch of the camera
  // Range is 0 to Math.PI radians
  public minPolarAngle = 0; // radians
  public maxPolarAngle = Math.PI; // radians

  public pointerSpeed = 1.5;
  public panSpeed = 1;
  public movementSpeed = 16;
  public shiftMultiplier = 0.45;
  public zoomScale = 100;
  public screenSpacePanning = true;

  private rotation = new Euler(0, 0, 0, 'YXZ');
  private look = new Vector3();

  private boundOnMouseMove = this.onMouseMove.bind(this);
  private boundOnMouseDown = this.onMouseDown.bind(this);
  private boundOnMouseUp = this.onMouseUp.bind(this);
  private boundOnKeyDown = this.onKeyDown.bind(this);
  private boundOnKeyUp = this.onKeyUp.bind(this);
  private boundOnScroll = this.onScroll.bind(this);

  private movement = {
    forward: 0,
    backward: 0,
    left: 0,
    right: 0,
    up: 0,
    down: 0,
    multiplier: 1,
  };

  constructor(private camera: Object3D, private domElement: HTMLElement) {
    super();

    this.connect();
  }

  connect() {
    this.domElement.addEventListener('mousemove', this.boundOnMouseMove);
    this.domElement.addEventListener('mousedown', this.boundOnMouseDown);
    this.domElement.addEventListener('mouseup', this.boundOnMouseUp);
    this.domElement.addEventListener('wheel', this.boundOnScroll, {
      passive: true,
    });
    window.addEventListener('keydown', this.boundOnKeyDown);
    window.addEventListener('keyup', this.boundOnKeyUp);
  }

  disconnect() {
    this.domElement.removeEventListener('mousemove', this.boundOnMouseMove);
    this.domElement.removeEventListener('mousedown', this.boundOnMouseDown);
    this.domElement.removeEventListener('mouseup', this.boundOnMouseUp);
    this.domElement.removeEventListener('wheel', this.boundOnScroll);
    window.removeEventListener('keydown', this.boundOnKeyDown);
    window.removeEventListener('keyup', this.boundOnKeyUp);
  }

  dispose() {
    this.disconnect();
  }

  update(dt: number) {
    if (this.movement.forward !== 0) {
      this.moveForward(this.movement.forward * dt * this.movement.multiplier);
    }

    if (this.movement.backward !== 0) {
      this.moveForward(-this.movement.backward * dt * this.movement.multiplier);
    }

    if (this.movement.right !== 0) {
      this.moveRight(this.movement.right * dt * this.movement.multiplier);
    }

    if (this.movement.left !== 0) {
      this.moveRight(-this.movement.left * dt * this.movement.multiplier);
    }

    if (this.movement.up !== 0) {
      this.moveUp(this.movement.up * dt * this.movement.multiplier);
    }

    if (this.movement.down !== 0) {
      this.moveUp(-this.movement.down * dt * this.movement.multiplier);
    }
  }

  getDirection(v: Vector3) {
    return v.set(0, 0, -1).applyQuaternion(this.camera.quaternion);
  }

  moveForward(distance: number) {
    const camera = this.camera;

    this.rotation.setFromQuaternion(camera.quaternion);
    this.look.setFromMatrixColumn(camera.matrix, 0);

    this.look.crossVectors(
      camera.up.clone().applyEuler(this.rotation),
      this.look
    );

    camera.position.addScaledVector(this.look, distance);
  }

  moveUp(distance: number) {
    const camera = this.camera;

    this.rotation.setFromQuaternion(camera.quaternion);
    this.look.copy(camera.up.clone().applyEuler(this.rotation));

    camera.position.addScaledVector(this.look, distance);
  }

  moveRight(distance: number) {
    const camera = this.camera;

    this.look.setFromMatrixColumn(camera.matrix, 0);

    camera.position.addScaledVector(this.look, distance);
  }

  private onMouseDown(event: MouseEvent) {
    if (event.button === 1) this.cameraPanning = true;
    if (event.button === 2) this.cameraMoving = true;
  }

  private onMouseUp(event: MouseEvent) {
    if (event.button === 1) this.cameraPanning = false;
    if (event.button === 2) this.cameraMoving = false;
  }

  private onScroll(event: WheelEvent) {
    if (event.deltaY < 0) {
      this.moveForward(this.movementSpeed * 0.002 * this.zoomScale);
    } else if (event.deltaY > 0) {
      this.moveForward(-this.movementSpeed * 0.002 * this.zoomScale);
    }
  }

  private rotate(event: MouseEvent) {
    const movementX = event.movementX || 0;
    const movementY = event.movementY || 0;

    const camera = this.camera;
    this.rotation.setFromQuaternion(camera.quaternion);

    this.rotation.y -= movementX * 0.002 * this.pointerSpeed;
    this.rotation.x -= movementY * 0.002 * this.pointerSpeed;

    this.rotation.x = Math.max(
      _PI_2 - this.maxPolarAngle,
      Math.min(_PI_2 - this.minPolarAngle, this.rotation.x)
    );

    camera.quaternion.setFromEuler(this.rotation);
  }

  private pan(event: MouseEvent) {
    const movementX = event.movementX || 0;
    const movementY = event.movementY || 0;

    const camera = this.camera;

    this.look.setFromMatrixColumn(camera.matrix, 0);
    camera.position.addScaledVector(
      this.look,
      -movementX * 0.02 * this.panSpeed
    );

    if (this.screenSpacePanning) {
      this.look.setFromMatrixColumn(camera.matrix, 1);
    } else {
      this.look.setFromMatrixColumn(camera.matrix, 0);
      this.look.crossVectors(camera.up, this.look);
    }

    camera.position.addScaledVector(
      this.look,
      movementY * 0.02 * this.panSpeed
    );
  }

  private onMouseMove(event: MouseEvent) {
    if (this.cameraMoving === false && this.cameraPanning === false) return;

    this.cameraMoving && this.rotate(event);
    this.cameraPanning && this.pan(event);

    this.dispatchEvent(_changeEvent);
  }

  private onKeyDown(event: KeyboardEvent) {
    if (!event.target || (event.target as HTMLElement).tagName !== 'BODY')
      return;

    if (event.altKey || event.ctrlKey) return;

    switch (event.code) {
      case 'ShiftLeft':
        this.movement.multiplier = this.shiftMultiplier;
        break;
      case 'KeyE':
        this.movement.up = this.movementSpeed;
        break;
      case 'KeyQ':
        this.movement.down = this.movementSpeed;
        break;
      case 'KeyW':
        this.movement.forward = this.movementSpeed;
        break;
      case 'KeyS':
        this.movement.backward = this.movementSpeed;
        break;
      case 'KeyA':
        this.movement.left = this.movementSpeed;
        break;
      case 'KeyD':
        this.movement.right = this.movementSpeed;
        break;
    }
  }

  private onKeyUp(event: KeyboardEvent) {
    if (!event.target || (event.target as HTMLElement).tagName !== 'BODY')
      return;

    if (event.altKey || event.ctrlKey) return;

    switch (event.code) {
      case 'ShiftLeft':
        this.movement.multiplier = 1;
        break;
      case 'KeyE':
        this.movement.up = 0;
        break;
      case 'KeyQ':
        this.movement.down = 0;
        break;
      case 'KeyW':
        this.movement.forward = 0;
        break;
      case 'KeyS':
        this.movement.backward = 0;
        break;
      case 'KeyA':
        this.movement.left = 0;
        break;
      case 'KeyD':
        this.movement.right = 0;
        break;
    }
  }
}

export { CameraControls };
