import { EngineComponent } from '../types/engine-component';
import { EngineEvents } from '../types/events';
import { GameRunner } from '../types/game-runner';
import { Instancable } from '../types/instancable';
import { EventEmitter } from '../utils/events';
import { Renderer } from './renderer';

export class Engine extends GameRunner {
  public lastTick = performance.now();
  public running = false;
  public events!: EventEmitter<EngineEvents>;
  public render!: Renderer;
  public element!: HTMLElement;
  public components: EngineComponent[] = [];

  override mount(element: HTMLElement): void {
    this.element = element;
    this.render = new Renderer(element);
    this.render.renderer.autoClear = false;
  }

  override loop(now: number): void {
    throw new Error('Method not implemented.');
  }

  override start(): void {
    this.running = true;
    this.loop(this.lastTick);
  }

  override stop(): void {
    this.running = false;
    for (const component of this.components) {
      component.dispose();
    }
    this.render.dispose();
  }

  /**
   * Update every component.
   */
  update(dt: number) {
    for (const component of this.components) {
      component.update(dt);
    }
  }

  /**
   * Use and initialize an engine component
   * @param Component Engine Component
   */
  use(Component: Instancable<EngineComponent>) {
    const component = new Component(this.render, this.events);
    this.components.push(component);
    component.initialize();
    return component;
  }

  /**
   * Get an engine component.
   * @param component Component
   */
  getComponent<T extends EngineComponent>(component: Instancable<T>) {
    return <T>this.components.find((cls) => cls.name === component.name);
  }
}
